/* wptJPG.cpp : JPG picture class
 *	Copyright (C) 2001 Dr.Yovav Gad <Sources@SuperMain.com>
 *	Copyright (C) 2005 Timo Schulz
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with PTD; if not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

/*-----------------------------------------------------------------------------
 * Picture (Implementations) Version 1.00
 *
 * Routinges for showing JPG pictur files
 *
 * Author: Dr. Yovav Gad, EMail: Sources@SuperMain.com ,Web: www.SuperMain.com
 * 
 * This version uses a stripped down version of Picture.cpp and Picture.h.
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>
#include <ocidl.h>
#include <olectl.h>

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>

#include "wptJPG.h"

#define HIMETRIC_INCH	2540
#define ERROR_TITLE	"CJPG Error"

#define out_of_core() do { \
	MessageBox (NULL, "Can not allocate memory", ERROR_TITLE, MB_OK|MB_ICONSTOP); \
	return FALSE; \
    } while (0)


CJPG::CJPG (void)
{
    m_IPicture = NULL;
    m_Height = 0;
    m_Weight = 0;
    m_Width = 0;
}


CJPG::~CJPG (void)
{
    if (m_IPicture != NULL)
	freePictureData (); 
}



/* Free the allocated memory that holdes the IPicture Interface data
   and clear picture information. */
void 
CJPG::freePictureData (void)
{
    if (m_IPicture != NULL) {
	m_IPicture->Release();
	m_IPicture = NULL;
	m_Height = 0;
	m_Weight = 0;
	m_Width = 0;	
    }
}


/* Open a JPG File And Load It Into IPicture (Interface) */
BOOL 
CJPG::load (LPCSTR sFilePathName)
{
    BOOL bResult = FALSE;
    FILE * f;
    int nSize = 0;

    if (m_IPicture != NULL) 
	freePictureData ();

    f = fopen (sFilePathName, "rb");
    if (f) {
	struct stat st;
	fstat (fileno (f), &st);
	nSize = st.st_size;
	BYTE *pBuffer = new BYTE[nSize];
	if  (!pBuffer) {
	    fclose (f);
	    out_of_core ();
	}
	memset (pBuffer, 0, nSize);
	if (fread(pBuffer, 1, nSize, f) > 0) {
	    if (loadPictureData (pBuffer, nSize))
		bResult = TRUE;
	}
	fclose (f);
	delete [] pBuffer;
    }
    else {
	MessageBox (NULL, strerror (errno), ERROR_TITLE, MB_OK|MB_ICONSTOP);
	bResult = FALSE;
    }

    m_Weight = nSize; /* Update Picture Size Info... */

    if(m_IPicture != NULL) {
	m_IPicture->get_Height (&m_Height);
	m_IPicture->get_Width (&m_Width);
	/* Calculate Its Size On a "Standard" (96 DPI) Device Context */
	m_Height = MulDiv (m_Height, 96, HIMETRIC_INCH);
	m_Width  = MulDiv (m_Width,  96, HIMETRIC_INCH);	
    }
    else {
	/* Picture data is not a known picture type */
	m_Height = 0;
	m_Width = 0;
	bResult = FALSE;	
    }
    return bResult;
}



/* read the picture data from a source (file / resource)
   and load it into the current IPicture object in use */
BOOL 
CJPG::loadPictureData (BYTE *pBuffer, int nSize)

{
    BOOL bResult = FALSE;
    HGLOBAL hGlobal;
    void* pData;
    IStream* pStream = NULL;
    
    hGlobal = GlobalAlloc (GMEM_MOVEABLE, nSize);
    if (hGlobal == NULL) {
	out_of_core ();
	return FALSE;
    }

    pData = GlobalLock (hGlobal);
    memcpy (pData, pBuffer, nSize);
    GlobalUnlock (hGlobal);

    if (CreateStreamOnHGlobal (hGlobal, TRUE, &pStream) == S_OK) {
	HRESULT hr;
	hr = OleLoadPicture (pStream, nSize, FALSE, IID_IPicture, 
			     (LPVOID *)&m_IPicture);
	if (hr == E_NOINTERFACE) {
	    MessageBox (NULL, "IPicture interface is not supported", 
			ERROR_TITLE, MB_OK|MB_ICONSTOP);
	    return FALSE;
	}
	else { /* S_OK */
	    pStream->Release ();
	    pStream = NULL;
	    bResult = TRUE;
	}	
    }

    GlobalFree (hGlobal);
    return (bResult);
}


/* Draw the loaded picture direct to the client DC */
BOOL 
CJPG::show (HDC pDC, POINT *leftTop, POINT *widthHeight, 
	    int magnifyX, int magnifyY)

{
    RECT drawRect;
    long width  = 0;
    long height = 0;
    HRESULT hrP = 0;

    if (pDC == NULL || m_IPicture == NULL) 
	return FALSE;

    m_IPicture->get_Width (&width);
    m_IPicture->get_Height (&height);
	
    if (magnifyX == 0) 
	magnifyX = 0;
    if (magnifyY == 0) 
	magnifyY = 0;
    magnifyX = int(MulDiv (width, GetDeviceCaps(pDC, LOGPIXELSX), HIMETRIC_INCH) * magnifyX);
    magnifyY = int(MulDiv (height, GetDeviceCaps(pDC, LOGPIXELSY), HIMETRIC_INCH) * magnifyY);

    drawRect.left = leftTop->x;
    drawRect.top = leftTop->y;
    drawRect.right = magnifyX;
    drawRect.bottom = magnifyY;

    hrP = m_IPicture->Render (pDC,
                      leftTop->x,               // Left
                      leftTop->y,               // Top
                      widthHeight->x +magnifyX, // Width
                      widthHeight->y +magnifyY, // Height
                      0,
                      height,
                      width,
                      -height,
                      &drawRect);

    if (SUCCEEDED (hrP)) 
	return (TRUE);

    out_of_core ();}


/* Get the original picture pixel size (ignore what current DC is using)
   pointer to a Device Context is needed for pixel calculation, */
BOOL 
CJPG::updateSizeOnDC (HDC pDC)

{	
    if(pDC == NULL || m_IPicture == NULL) { 
	m_Height = 0; 
	m_Width = 0; 
	return (FALSE); 
    }

    m_IPicture->get_Height (&m_Height);
    m_IPicture->get_Width (&m_Width);

    /* Get Current DPI - Dot Per Inch */
    int CurrentDPI_X = GetDeviceCaps (pDC, LOGPIXELSX);
    int CurrentDPI_Y = GetDeviceCaps (pDC, LOGPIXELSY);

    m_Height = MulDiv (m_Height, CurrentDPI_Y, HIMETRIC_INCH);
    m_Width  = MulDiv (m_Width,  CurrentDPI_X, HIMETRIC_INCH);

    return (TRUE);
}

/* Return height of the current image. */
LONG
CJPG::getHeight (void)
{
    return m_Height;
}


/* Return weight of the current image. */
LONG 
CJPG::getWeight (void)
{
    return m_Weight;
}


/* Return width of the current image. */
LONG
CJPG::getWidth (void)
{
    return m_Width;
}
