/* wptKeygenDlg.cpp - Key Generation dialog
 *	Copyright (C) 2000-2005 Timo Schulz
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with WinPT; if not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>

#include "resource.h"
#include "wptTypes.h"
#include "wptNLS.h"
#include "wptGPG.h"
#include "wptCommonCtl.h"
#include "wptContext.h" /* for passphrase_s */
#include "wptDlgs.h"
#include "wptW32API.h"
#include "wptVersion.h"
#include "wptErrors.h"
#include "wptUTF8.h"

/* All valid key generation combination. */
enum gpg_keytype_t {
    GPG_KEYGEN_NONE    = 0,
    GPG_KEYGEN_DSA_ELG = 1,
    GPG_KEYGEN_DSA_RSA = 2,
    GPG_KEYGEN_DSA_SIG = 3,
    GPG_KEYGEN_RSA_SIG = 4,    
    GPG_KEYGEN_RSA     = 5,
    GPG_KEYGEN_RSA_RSA = 6 /*PGP*/
};


static const char key_params[] =
	"<GnupgKeyParms format=\"internal\">\n"
	"Key-Type: %s\n"
	"Key-Usage: sign\n"
	"Key-Length: %d\n"
	"Subkey-Type: %s\n"
	"Subkey-Length: %d\n"
	"Name-Real: %s\n"	
	"Name-Email: %s\n"
	"Expire-Date: %s\n"
	"Passphrase: %s\n"
	"</GnupgKeyParms>\n";

static const char key_params_with_comment[] = 
	"<GnupgKeyParms format=\"internal\">\n"
	"Key-Type: %s\n"
	"Key-Usage: sign\n"
	"Key-Length: %d\n"
	"Subkey-Type: %s\n"
	"Subkey-Length: %d\n"
	"Name-Real: %s\n"
	"Name-Comment: %s\n"
	"Name-Email: %s\n"
	"Expire-Date: %s\n"
	"Passphrase: %s\n"
	"</GnupgKeyParms>\n";

static const char key_params_one[] = 
	"<GnupgKeyParms format=\"internal\">\n"
	"Key-Type: %s\n"
	"Key-Length: %d\n"
	"Key-Usage: %s\n"	
	"Name-Real: %s\n"	
	"Name-Email: %s\n"
	"Expire-Date: %s\n"
	"Passphrase: %s\n"
	"</GnupgKeyParms>\n";

static const char key_params_one_with_comment[] = 
	"<GnupgKeyParms format=\"internal\">\n"
	"Key-Type: %s\n"
	"Key-Length: %d\n"
	"Key-Usage: %s\n"	
	"Name-Real: %s\n"	
	"Name-Email: %s\n"
	"Name-Comment: %s\n"
	"Expire-Date: %s\n"
	"Passphrase: %s\n"
	"</GnupgKeyParms>\n";


/* Generate the GPG specific genkey params with the given information.
   @keytype: See valid combinations.
   @bits: Length in bits.
   @user: user-ID name
   @comment: comment for the user-ID.
   @email: email address.
   @expdata: date of expiration or NULL.
   @passphrase: the actual passphrase.
   Return value: the gen. params. */
static char*
gpg_genkey_params (int keytype, int bits, 
                   const char *user, const char *comment, const char *email,
                   const char *expdate, const char *passphrase)
{
    char *p = NULL;
    int addsize = strlen ("sign encrypt");
    int size = 0;
    
    if (keytype == GPG_KEYGEN_NONE)
        return NULL;
    
    size = strlen (user) + strlen (email) + strlen (passphrase) + addsize + 48;
    if (comment && *comment)
        size += strlen (key_params_with_comment) + strlen (comment);
    else
        size += strlen (key_params);
    if (expdate)
        size += strlen (expdate) + 1;
    p = (char *)malloc (size+1);
    if (!p)
        BUG (0);
    /* XXX: simply the whole switch-case code. */
    if (comment && *comment) {
        switch( keytype ) {
        case GPG_KEYGEN_DSA_ELG:
            sprintf (p, key_params_with_comment,
                     "DSA", 1024, "ELG-E", bits, user, comment, email,
                     expdate ? expdate : "0", passphrase);
            break;
            
        case GPG_KEYGEN_DSA_RSA:
            sprintf( p, key_params_with_comment, 
                     "DSA", 1024, "RSA", bits, user, comment, email,
                     expdate ? expdate : "0", passphrase );
            break;
            
        case GPG_KEYGEN_DSA_SIG:
            sprintf( p, key_params_one_with_comment, 
                     "DSA", 1024, "sign",
                     user, comment, email,
                     expdate ? expdate : "0", passphrase );
            break;
            
        case GPG_KEYGEN_RSA_SIG:
            sprintf( p, key_params_one_with_comment, 
                     "RSA", bits, "sign",
                     user, comment, email,
                     expdate ? expdate : "0", passphrase );
            break;
            
        case GPG_KEYGEN_RSA:
            sprintf( p, key_params_one_with_comment, 
                    "RSA", bits, "sign encrypt",
                    user, comment, email,
                    expdate ? expdate : "0", passphrase );
            break;

	case GPG_KEYGEN_RSA_RSA:
	    sprintf( p, key_params_with_comment,
		     "RSA", bits, "RSA", bits, user, comment, email,
		     expdate? expdate : "0", passphrase );
	    break;
            
        default:
	    if (p)
		free (p);
	    p = NULL;
            break;
        }
    }
    else {
        switch ( keytype ) {
        case GPG_KEYGEN_DSA_ELG:
            sprintf( p, key_params,
                    "DSA", 1024, "ELG-E", bits, user, email,
                    expdate ? expdate : "0", passphrase );
            break;
            
        case GPG_KEYGEN_DSA_RSA:
            sprintf( p, key_params,
                    "DSA", 1024, "RSA", bits, user, email,
                    expdate ? expdate : "0", passphrase );
            break;
            
        case GPG_KEYGEN_DSA_SIG:
            sprintf( p, key_params_one, 
                    "DSA", 1024, "sign",
                    user, email,
                    expdate ? expdate : "0", passphrase );
            break;
            
        case GPG_KEYGEN_RSA_SIG:
            sprintf( p, key_params_one, 
                    "RSA", bits, "sign",
                    user, email,
                    expdate ? expdate : "0", passphrase );
            break;
            
        case GPG_KEYGEN_RSA:
            sprintf( p, key_params_one, 
                    "RSA", bits, "sign encrypt",
                    user, email,
                    expdate ? expdate : "0", passphrase );
            break;

	case GPG_KEYGEN_RSA_RSA:
	    sprintf( p, key_params,
		     "RSA", bits, "RSA", bits, user, email,
		     expdate? expdate : "0", passphrase );
	    break;
            
        default:
	    if (p)
		free (p);
	    p = NULL;
            break;
        }
    }
    return p;
}


/* Generate a key with the given params @params. @prog_cb is a user defined
   progress callback which is called during the generation.
   @fpr will store the fingerprint of the generated key.
   Return value: 0 on success. */
gpgme_error_t
gpg_genkey (const char *params, gpgme_progress_cb_t prog_cb, char **fpr)
{
    gpgme_error_t err = 0;
    gpgme_ctx_t ctx;
    gpgme_genkey_result_t res;
    
    err = gpgme_new(&ctx);
    if (err)
        return err;
    if (prog_cb)
        gpgme_set_progress_cb (ctx, prog_cb, NULL);
    err = gpgme_op_genkey (ctx, params, NULL, NULL);
    if (!err) {
	res = gpgme_op_genkey_result (ctx);
	*fpr = res->fpr? strdup (res->fpr) : NULL;
    }
    gpgme_release (ctx);
    return err;
}



/* Clear all dialog fields. */
static void 
clear_dlg_fields (HWND dlg)
{
    SetDlgItemText (dlg, IDC_KEYGEN_SUBKEYBITS, "");
    SetDlgItemText (dlg, IDC_KEYGEN_NAME, "");
    SetDlgItemText (dlg, IDC_KEYGEN_EMAIL, "");
    SetDlgItemText (dlg, IDC_KEYGEN_COMMENT, "");
    SetDlgItemText (dlg, IDC_KEYGEN_EXPDATE, "");
    SetDlgItemText (dlg, IDC_KEYGEN_PASSPHRASE, "");
    SetDlgItemText (dlg, IDC_KEYGEN_PWDCHECK, "");
}


/* Ask the user if a keyring backup is wanted and if so,
   backup both keyrings to the selected folder. @dlg is
   the handle of the parent window.*/
static void
backup_keyrings (HWND dlg)
{
    const char *name;
    char *path = NULL;
    char *keyring = NULL;
    int id;
    
    path = get_gnupg_path ();
    if (!path)
	BUG (dlg);
    id = msg_box (dlg, 
                 _("It is STRONGLY recommend that you backup your keyrings because they both "
                   "contain VERY important data.\nRemember that your hard disk can crash or the "
                   "files can be deleted by accident; so it is a good\nidea to store them on "
                   "a different mass stoarge like a floppy or CDR!\n\n"
                   "Backup your keyrings now?"),
                 _("WARNING - Important hint" ), MB_YESNO);
    if (id == IDYES) {
        name = get_filesave_dlg (dlg, _("Destination for Public Keyring"), 
				 NULL, "pubring.gpg");
        if( name ) {
	    keyring = make_filename (path, "pubring", "gpg");
            if( !CopyFile (keyring, name, FALSE))
		log_box (_("Key Generation"), MB_ERR,
		         _("Could not copy %s -> %s"), keyring, name);
	    free_if_alloc (keyring);
        }
        name = get_filesave_dlg (dlg, _("Destination for Secret Keyring"), 
				 NULL, "secring.gpg");
        if( name ) {
	    keyring = make_filename (path, "secring", "gpg");
            if (!CopyFile (keyring, name, FALSE))
		log_box (_("Key Generation"), MB_ERR,
			 _("Could not copy %s -> %s"), keyring, name);
	    free_if_alloc (keyring);
        }
    }
    free_if_alloc (path);
}


/* Fill in all valid GPG algorithms. */
static void
fill_keytype_box (HWND dlg)
{
    HWND cb = GetDlgItem (dlg, IDC_KEYGEN_KEYTYPE);

#define addstr(cb, str) \
    SendMessage ((cb), CB_ADDSTRING, 0, (LPARAM)(LPCTSTR)(str))
    addstr (cb, _("DSA and ELG (default)"));
    addstr (cb, _("DSA and RSA"));
    addstr (cb, _("DSA sign only"));
    addstr (cb, _("RSA sign only"));
    addstr (cb, _("RSA sign and encrypt"));
    addstr (cb, _("RSA and RSA (PGP)") );
    SendMessage (cb, CB_SETCURSEL, 0, 0);
#undef addstr
}


/* Check that the given date lies not in the past.
   Return value: 1 on success. */
int
keygen_check_date (SYSTEMTIME *st)
{	
    SYSTEMTIME t;
	
    GetSystemTime (&t);
    if (st->wYear > t.wYear || st->wMonth > t.wMonth)
	return 1;
    else if (st->wYear < t.wYear || st->wMonth < t.wMonth || st->wDay < t.wDay)
	return 0;
    return 1;
}


/* Dialog box procedure for key generation. */
BOOL CALLBACK
keygen_dlg_proc (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam)
{
    static genkey_s *ctx;
    SYSTEMTIME st;
    gpgme_error_t err;
    char name[128] = {0}, email[128] = {0}, comment[128] = {0};
    char *pwd;
    char t[64], *expire = NULL, *fpr=NULL;
    int bits, use_comment, keytype = 0;
    int cancel = 0;
    char *p;
    
    switch ( msg ) {
    case WM_INITDIALOG:
	if (lparam != 0)
	    ctx = (genkey_s *)lparam;
        SetWindowText (dlg, _("Key Generation"));
        SetDlgItemText(dlg, IDC_KEYGEN_INFO,
			_("NOTE: Key generation can be a lengthy process! Please wait until "
			  "you get the message that key generation was finished."));
        SetDlgItemText (dlg, IDC_KEYGEN_SUBKEYINF, _("Subkey size in &bits"));
        SetDlgItemText (dlg, IDC_KEYGEN_NAMEINF, _("&Real name"));
        SetDlgItemText (dlg, IDC_KEYGEN_COMMINF, _("&Comment (optional)"));
        SetDlgItemText (dlg, IDC_KEYGEN_EMAILINF, _("Email &address"));
        SetDlgItemText (dlg, IDC_KEYGEN_EXPINF, _("&Expire date"));
        SetDlgItemText (dlg, IDC_KEYGEN_KEYTYPEINF, _("Key &type"));
	SetDlgItemText (dlg, IDC_KEYGEN_EXPNEVER, _("&Never"));
	SetDlgItemText (dlg, IDCANCEL, _("&Cancel"));

        SetDlgItemInt (dlg, IDC_KEYGEN_SUBKEYBITS, DFAULT_KEYSIZE, FALSE);
	CheckDlgButton (dlg, IDC_KEYGEN_HIDEPWD, BST_CHECKED);
	CheckDlgButton (dlg, IDC_KEYGEN_EXPNEVER, BST_CHECKED);
	EnableWindow (GetDlgItem (dlg, IDC_KEYGEN_EXPDATE), FALSE);
        fill_keytype_box (dlg);
	center_window (dlg, NULL);
        SetForegroundWindow (dlg);
        return TRUE;
        
    case WM_SYSCOMMAND:
        if (LOWORD (wparam) == SC_CLOSE)
            EndDialog (dlg, TRUE);
        return FALSE;
        
    case WM_COMMAND:
	if (HIWORD (wparam) == BN_CLICKED &&
	    LOWORD (wparam) == IDC_KEYGEN_EXPNEVER) {
	    int never = IsDlgButtonChecked (dlg, IDC_KEYGEN_EXPNEVER);
	    EnableWindow (GetDlgItem (dlg, IDC_KEYGEN_EXPDATE), !never);
	}

        switch (LOWORD (wparam)) {
        case IDOK:
            bits = GetDlgItemInt (dlg, IDC_KEYGEN_SUBKEYBITS, NULL, FALSE);
            if (bits < 1024 || bits > 4096) {
		msg_box (dlg, _("Invalid value. Allowed values 1024-4096 bits."), 
			 _("Key Generation"), MB_ERR);
                return FALSE;
            }
            if (bits > DFAULT_KEYSIZE) {
                int id = msg_box (dlg, _("Do you really need such a large key?"),
				  _("Key Generation"), MB_YESNO);
                if (id == IDNO)
                    bits = DFAULT_KEYSIZE;
            }
            if (!GetDlgItemText (dlg, IDC_KEYGEN_NAME, name, sizeof (name) - 1)) {
                msg_box( dlg, _("Please enter the name."), _("Key Generation"), MB_ERR);
                return FALSE;
            }
	    if (strchr (name, '@')) {
		msg_box (dlg, _("Please do not enter the email address in the name field."),
			 _("Key Generation"), MB_INFO);
		return FALSE;
	    }
            if (!GetDlgItemText (dlg, IDC_KEYGEN_EMAIL, email, sizeof (email) -1)
                || check_email_address (email)) {
                msg_box (dlg, _("Please enter a valid email address."),
			 _("Key Generation"), MB_ERR);
                return FALSE;
            }
            use_comment = GetDlgItemText(dlg, IDC_KEYGEN_COMMENT, 
					 comment, sizeof (comment) -1);
	    if (use_comment > 0 && strchr (comment, '@')) {
		msg_box (dlg, _("Please do NOT enter the email address in the comment field."),
			 _("Key Generation"), MB_INFO);
		return FALSE;
	    }
            keytype = SendDlgItemMessage (dlg, IDC_KEYGEN_KEYTYPE, CB_GETCURSEL, 0, 0) + 1;
            if (IsDlgButtonChecked (dlg, IDC_KEYGEN_EXPNEVER))
                expire = NULL;
            else {
		DateTime_GetSystemtime (GetDlgItem (dlg, IDC_KEYGEN_EXPDATE), &st);
		_snprintf (t, DIM (t)-1, "%04d-%02d-%02d", st.wYear, st.wMonth, st.wDay);
                expire = t;
	    }

            pwd = request_passphrase2 (_("Key Generation"), PASSDLG_STRICT, &cancel);
	    if (cancel) {
		sfree_if_alloc (pwd);
		return FALSE;
	    }
	    if (!pwd) {
                msg_box (dlg, _("Please enter the passphrase."), 
			 _("Key Generation"), MB_ERR);
                return FALSE;
            }

            if (!use_comment && !strlen (comment)) {
                char *utf8_name;
		utf8_name = native_to_utf8 (name);
		if (!utf8_name)
		    BUG (0);
                p = gpg_genkey_params (keytype, bits, utf8_name, NULL, 
				       email, expire, pwd);
                safe_free (utf8_name);
            }
            else {
                char *utf8_name, *utf8_comment;
		utf8_name = native_to_utf8 (name);
		utf8_comment = native_to_utf8 (comment);
		if( !utf8_name || !utf8_comment )
		    BUG (0);
                p = gpg_genkey_params (keytype, bits, utf8_name, utf8_comment, 
				       email, expire, pwd);
                safe_free (utf8_name);
                safe_free (utf8_comment);
            }
            keygen_cb_dlg_create ();
            err = gpg_genkey (p, keygen_cb, &fpr);
	    sfree_if_alloc (pwd);
            if (p) {
		wipememory (p, strlen (p));  /* burn the passphrase! */
                free (p);
            }
            keygen_cb_dlg_destroy ();
            keygen_cb (NULL, NULL, 0, 0, 0); /* flush */
            if (err) {
		safe_free (fpr);
                msg_box (dlg, gpgme_strerror (err), _("Key Generation"), MB_ERR);
                return FALSE;
            }
            status_box (dlg, _("Key Generation completed"), _("GnuPG Status"));
            
	    keycache_update (0, fpr);
	    keycache_update (1, fpr);
	    safe_free (fpr);	    

            clear_dlg_fields (dlg);
            backup_keyrings (dlg);
	    if (ctx)
		ctx->cancel = 0;
            EndDialog (dlg, TRUE);
            return TRUE;
            
        case IDCANCEL:
	    if (ctx)
		ctx->cancel = 1;
            EndDialog (dlg, FALSE);	   
            return FALSE;
        }
        break;
    }
    
    return FALSE;
}


BOOL CALLBACK
keygen_wizard_dlg_proc (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam)
{
    static genkey_s *ctx;
    static int pubkey_algo = GPG_KEYGEN_DSA_ELG;
    gpgme_error_t err;
    char name[128], email[128];
    char * utf8_name, * p, *fpr=NULL;
    char * pass = NULL;
    int cancel = 0;
    

    switch( msg ) {
    case WM_INITDIALOG:
	ctx = (genkey_s *)lparam;
	if (!ctx || (ctx && ctx->interactive == 0))
	    EnableWindow (GetDlgItem (dlg, IDC_KEYWIZARD_EXPERT), FALSE);
	SetDlgItemText (dlg, IDC_KEYWIZARD_USERSA, _("&Prefer RSA keys"));
	SetDlgItemText (dlg, IDC_KEYWIZARD_NAMEINF, _("Real name:"));
	SetDlgItemText (dlg, IDC_KEYWIZARD_EMAILINF, _("Email address:"));
	SetDlgItemText (dlg, IDC_KEYWIZARD_TITLEINF, _("Name and E-Mail Assignment"));
	SetDlgItemText (dlg, IDC_KEYWIZARD_TEXT1INF, _("Every key pair must have a name associated with it. The name and\nemail address let your correspondents that your public key they are\nusing belongs to us."));
	SetDlgItemText (dlg, IDC_KEYWIZARD_TEXT2INF, _("By accosiating an email address with your key pair, you will enable WinPT to assist your correspondents in selecting the correct public\nkey when communicating with you."));
	SetWindowText (dlg, _("Key Generation Wizard"));
	SetDlgItemText (dlg, IDCANCEL, _("&Cancel"));
	SetDlgItemText (dlg, IDC_KEYWIZARD_EXPERT, _("E&xpert"));
	SetForegroundWindow (dlg);
	center_window (dlg, NULL);
	break;

    case WM_SYSCOMMAND:
	if( LOWORD( wparam ) == SC_CLOSE )
	    EndDialog( dlg, FALSE );

    case WM_COMMAND:
	switch( LOWORD( wparam ) ) {
	case IDC_KEYWIZARD_EXPERT:
	    DialogBoxParam (glob_hinst, (LPCTSTR)IDD_WINPT_KEYGEN, dlg,
			    keygen_dlg_proc, 0);
	    EndDialog (dlg, TRUE);
	    break;

	case IDOK:
	    if( !GetDlgItemText( dlg, IDC_KEYWIZARD_NAME, name, sizeof name-1 ) ) {
		msg_box( dlg, _("Please enter the name."), 
			 _("Key Generation Wizard"), MB_ERR );
		return FALSE;
	    }
	    if (strchr (name, '@')) {
		msg_box (dlg, _("Please do not enter the email address in the name field."),
			 _("Key Generation Wizard"), MB_WARN);
		return FALSE;
	    }
	    if (!GetDlgItemText(dlg, IDC_KEYWIZARD_EMAIL, email, sizeof email-1 )
		|| check_email_address (email)) { 
		msg_box (dlg, _("Please enter a valid email address."), 
		         _("Key Generation Wizard"), MB_ERR);
		return FALSE;
	    }
	    if (strchr (email, '<') || strchr (email, '>')) {
		msg_box (dlg, _("Please do not add '<' or '>' to the email address."),
			 _("Key Generation Wizard"), MB_WARN);
		return FALSE;
	    }
	    pass = request_passphrase2 (_("Key Generation"), PASSDLG_STRICT, &cancel);
	    if (cancel)
		return FALSE;
	    utf8_name = native_to_utf8 (name);
	    if (!utf8_name)
		BUG (0);
	    if (IsDlgButtonChecked (dlg, IDC_KEYWIZARD_USERSA))
		pubkey_algo = GPG_KEYGEN_DSA_RSA;
	    p = gpg_genkey_params (pubkey_algo, DFAULT_KEYSIZE, utf8_name, 
				     NULL, email, NULL, pass);
	    free( utf8_name );
	    keygen_cb_dlg_create();
	    err = gpg_genkey (p, keygen_cb, &fpr);
	    keygen_cb_dlg_destroy();
	    keygen_cb( NULL, NULL, 0, 0, 0 );
	    if( p ) {
		memset( p, 0, strlen( p ) );
		free( p );
	    }
	    sfree_if_alloc (pass);
	    if( err ) {
                msg_box( dlg, gpgme_strerror( err ), _("Key Generation Wizard"), MB_ERR );
		if (fpr)
		    free (fpr);
                return FALSE;
            }
            status_box( dlg, _("Key Generation completed"), _("GnuPG Status") );
            
	    keycache_update (0, fpr);
	    keycache_update (1, fpr);
	    safe_free (fpr);

	    backup_keyrings (dlg);
	    if (ctx)
		ctx->cancel = 0;	    
	    EndDialog (dlg, TRUE);
	    break;

	case IDCANCEL:
	    if (ctx)
		ctx->cancel = 1;
	    EndDialog (dlg, FALSE);
	    break;
	}
	break;
    }
    return FALSE;
}
